@@ -47,15 +47,21 @@ class MicAudio(object):
     BLOCK_SIZE_SAMPLES = int(SAMPLE_RATE / float(BLOCKS_PER_SECOND))  # Block size in number of samples
     BLOCK_DURATION_MS = int(1000 * BLOCK_SIZE_SAMPLES // SAMPLE_RATE)  # Block duration in milliseconds
 
-    def __init__(self, callback=None, buffer_s=0, flush_queue=True, start=True, input_device_index=None, self_threaded=None):
+    def __init__(self, callback=None, buffer_s=0, flush_queue=True, start=True, input_device=None, self_threaded=None, reconnect_callback=None):
         self.callback = callback if callback is not None else lambda in_data: self.buffer_queue.put(in_data, block=False)
-        self.flush_queue = flush_queue
-        self.input_device_index = int(input_device_index) if input_device_index is not None else None
+        self.flush_queue = bool(flush_queue)
+        self.input_device = input_device
         self.self_threaded = bool(self_threaded)
+        if reconnect_callback is not None and not callable(reconnect_callback):
+            _log.error("Invalid reconnect_callback not callable: %r", reconnect_callback)
+            reconnect_callback = None
+        self.reconnect_callback = reconnect_callback
 
         self.buffer_queue = queue.Queue(maxsize=(buffer_s * 1000 // self.BLOCK_DURATION_MS))
         self.stream = None
         self.thread = None
+        self.thread_cancelled = False
+        self.device_info = None
         self._connect(start=start)
 
     def _connect(self, start=None):
@@ -69,11 +75,12 @@ class MicAudio(object):
             dtype=self.FORMAT,
             blocksize=self.BLOCK_SIZE_SAMPLES,
             # latency=80,
-            device=self.input_device_index,
+            device=self.input_device,
             callback=proxy_callback if not self.self_threaded else None,
         )
 
         if self.self_threaded:
+            self.thread_cancelled = False
             self.thread = threading.Thread(target=self._reader_thread, args=(callback,))
             self.thread.daemon = True
             self.thread.start()
@@ -85,11 +92,11 @@ class MicAudio(object):
         hostapi_info = sounddevice.query_hostapis(device_info['hostapi'])
         _log.info("streaming audio from '%s' using %s: %i sample_rate, %i block_duration_ms, %i latency_ms",
             device_info['name'], hostapi_info['name'], self.stream.samplerate, self.BLOCK_DURATION_MS, int(self.stream.latency*1000))
+        self.device_info = device_info
 
     def _reader_thread(self, callback):
-        while self.stream and not self.stream.closed:
-            read_available = self.stream.read_available
-            if read_available >= self.stream.blocksize:
+        while not self.thread_cancelled and self.stream and not self.stream.closed:
+            if self.stream.active and self.stream.read_available >= self.stream.blocksize:
                 in_data, overflowed = self.stream.read(self.stream.blocksize)
                 # print('_reader_thread', read_available, len(in_data), overflowed, self.stream.blocksize)
                 if overflowed:
@@ -102,11 +109,18 @@ class MicAudio(object):
         self.stream.close()
 
     def reconnect(self):
-        self.stream.close()
+        # FIXME: flapping
+        old_device_info = self.device_info
+        self.thread_cancelled = True
         if self.thread:
             self.thread.join()
             self.thread = None
+        self.stream.close()
         self._connect(start=True)
+        if self.reconnect_callback is not None:
+            self.reconnect_callback(self)
+        if self.device_info != old_device_info:
+            raise EngineError("Audio reconnect could not reconnect to the same device")
 
     def start(self):
         self.stream.start()
@@ -303,9 +317,11 @@ class AudioStore(object):
     - *save_dir* (*str*, default *None*): if set, the directory to save the `retain.tsv` file and optionally wav files.
     - *save_metadata* (*bool*, default *None*): whether to automatically save the recognition metadata.
     - *save_audio* (*bool*, default *None*): whether to automatically save the recognition audio data (in addition to just the recognition metadata).
+    - *retain_approval_func* (*Callable*, default *None*): if set, will be called with the `AudioStoreEntry` object about to be saved,
+        and should return `bool` whether to actually save. Example: `retain_approval_func=lambda entry: bool(entry.grammar_name != 'noisegrammar')`
     """
 
-    def __init__(self, audio_obj, maxlen=None, save_dir=None, save_audio=None, save_metadata=None, auto_save_predicate_func=None):
+    def __init__(self, audio_obj, maxlen=None, save_dir=None, save_audio=None, save_metadata=None, retain_approval_func=None):
         self.audio_obj = audio_obj
         self.maxlen = maxlen
         self.save_dir = save_dir
@@ -313,7 +329,7 @@ class AudioStore(object):
         self.save_metadata = save_metadata
         if self.save_dir:
             _log.info("retaining recognition audio and/or metadata to '%s'", self.save_dir)
-        self.auto_save_predicate_func = auto_save_predicate_func
+        self.retain_approval_func = retain_approval_func
         self.deque = collections.deque(maxlen=maxlen) if maxlen else None
         self.blocks = []
 
@@ -330,8 +346,6 @@ class AudioStore(object):
             if len(self.deque) == self.deque.maxlen:
                 self.save(-1)  # Save oldest, which is about to be evicted
             self.deque.appendleft(entry)
-        # if self.auto_save_predicate_func and self.auto_save_predicate_func(*entry):
-        #     self.save(0)
         self.blocks = []
 
     def cancel(self):
@@ -348,7 +362,9 @@ class AudioStore(object):
             return
 
         entry = self.deque[index]
-        if not self.save_audio and not self.save_metadata and not entry.force_save:
+        if (not self.save_audio) and (not self.save_metadata) and (not entry.force_save):
+            return
+        if self.retain_approval_func and not self.retain_approval_func(entry):
             return
         if self.save_audio or entry.force_save:
             filename = os.path.join(self.save_dir, "retain_%s.wav" % datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S_%f"))
@@ -397,7 +413,7 @@ class AudioStoreEntry(object):
         self.force_save = force_save
 
     def set(self, key, value):
-        """ Sets given key (as *str*) to value, returning the AudioStoreEntry for chaining; usable in `lambda`s. """
+        """ Sets given key (as *str*) to value, returning the AudioStoreEntry for chaining; usable in lambda functions. """
         setattr(self, key, value)
         return self
 
